@@ -1,3 +1,4 @@
+import logging
 from datetime import datetime, timedelta
 from django.conf import settings
 from django.shortcuts import render
@@ -10,33 +11,25 @@ import phonenumbers
 from twilio.base.exceptions import TwilioRestException
 from .forms import *
 from .utils import *
-from .conf import *
 from .dispatch import *
 
 
 __all__ = [
-    "Twilio2FAIndexView", "Twilio2FARegisterView", "Twilio2FAChangeView", "Twilio2FAStartView", "Twilio2FAVerifyView", "Twilio2FASuccessView",
-    "Twilio2FAFailedView",
+    "Twilio2FAIndexView", "Twilio2FARegisterView", "Twilio2FAChangeView", "Twilio2FAStartView", "Twilio2FAVerifyView",
+    "Twilio2FASuccessView", "Twilio2FAFailedView",
 ]
 
 
+logger = logging.getLogger("django_twilio_2fa")
+
+
 class Twilio2FAMixin(object):
-    SESSION_PREFIX = "twilio_2fa_"
-    URL_PREFIX = "twilio_2fa:"
-
-    SESSION_SID = "sid"
-    SESSION_TIMESTAMP = "timestamp"
-    SESSION_METHOD = "method"
-    SESSION_CAN_RETRY = "can_retry"
-
     # Session values that should be cleared
     SESSION_VALUES = [
         SESSION_SID, SESSION_TIMESTAMP, SESSION_METHOD,
-        SESSION_CAN_RETRY
+        SESSION_CAN_RETRY, SESSION_ATTEMPTS,
     ]
 
-    DATEFMT = "%Y%m%d%H%M%S"
-
     AVAILABLE_METHODS = {
         "sms": {
             "value": "sms",
@@ -98,8 +91,14 @@ class Twilio2FAMixin(object):
         else:
             self.allowed_methods = []
 
+        if request.GET.get("next"):
+            self.set_session_value(
+                SESSION_NEXT_URL,
+                request.GET.get("next")
+            )
+
     def dispatch(self, request, *args, **kwargs):
-        view_name = request.resolver_match.view_name.replace(self.URL_PREFIX, "")
+        view_name = request.resolver_match.view_name.replace(URL_PREFIX, "")
 
         if view_name in ["failed"]:
             # Always allow these views to be dispatched properly
@@ -124,12 +123,13 @@ class Twilio2FAMixin(object):
         return ctx
 
     def get_redirect(self, view_name, *args, **kwargs):
+        logger.debug(f"Redirecting to: {view_name}")
         return HttpResponseRedirect(
-            reverse(f"{self.URL_PREFIX}{view_name}", args=args, kwargs=kwargs)
+            reverse(f"{URL_PREFIX}:{view_name}", args=args, kwargs=kwargs)
         )
 
     def get_error_redirect(self, can_retry=False):
-        self.set_session_value(self.SESSION_CAN_RETRY, can_retry)
+        self.set_session_value(SESSION_CAN_RETRY, can_retry)
         return self.get_redirect("failed")
 
     def handle_twilio_exception(self, exc):
@@ -144,12 +144,17 @@ class Twilio2FAMixin(object):
         raise
 
     def get_session_value(self, key, default=None):
-        key = f"{self.SESSION_PREFIX}_{key}"
+        key = f"{SESSION_PREFIX}_{key}"
         return self.request.session.get(key, default)
 
     def set_session_value(self, key, value):
-        key = f"{self.SESSION_PREFIX}_{key}"
+        key = f"{SESSION_PREFIX}_{key}"
         self.request.session[key] = value
+        return value
+
+    def clear_session(self):
+        for key in self.SESSION_VALUES:
+            del self.request.session[f"{SESSION_PREFIX}_{key}"]
 
 
 class Twilio2FAVerificationMixin(Twilio2FAMixin):
@@ -173,6 +178,18 @@ class Twilio2FAVerificationMixin(Twilio2FAMixin):
             self.phone_number = None
 
     def dispatch(self, request, *args, **kwargs):
+        timeout_value = self.get_session_value(SESSION_TIMEOUT)
+
+        if timeout_value:
+            timeout_value = datetime.strptime(timeout_value, DATEFMT)
+
+            if timeout_value > datetime.now(tz=timeout_value.tzinfo):
+                messages.error(
+                    request,
+                    "You cannot make another 2FA verification attempt at this time."
+                )
+                return self.get_error_redirect(can_retry=False)
+
         if not self.phone_number:
             messages.warning(
                 request,
@@ -238,11 +255,20 @@ class Twilio2FAVerificationMixin(Twilio2FAMixin):
         return n[:-4] + self.e164_phone_number()[-4:]
 
     def update_verification_status(self, status):
-        twilio_sid = self.get_session_value(self.SESSION_SID)
+        twilio_sid = self.get_session_value(SESSION_SID)
 
         if not twilio_sid:
             return True
 
+        twilio_2fa_verification_status_changed.send(
+            None,
+            user=self.request.user,
+            phone_number=self.e164_phone_number(),
+            method=self.get_session_value(SESSION_METHOD),
+            twilio_sid=twilio_sid,
+            status=status
+        )
+
         try:
             (get_twilio_client().verify
                 .services(get_setting("SERVICE_ID"))
@@ -363,7 +389,7 @@ class Twilio2FAStartView(Twilio2FAVerificationMixin, TemplateView):
         action = request.GET.get("action")
 
         if not self.phone_number:
-            return self.get_redirect("start")
+            return self.get_redirect("register")
 
         elif action and action == "retry":
             r = self.retry_action(request, *args, **kwargs)
@@ -371,7 +397,10 @@ class Twilio2FAStartView(Twilio2FAVerificationMixin, TemplateView):
             if r is not None:
                 return r
 
-        elif request.method == "GET" and len(self.allowed_methods) == 1:
+        if request.method == "GET":
+            self.clear_session()
+
+        if request.method == "GET" and len(self.allowed_methods) == 1:
             # If only one option exists, we start the verification and send the user on
             self.send_verification(
                 self.allowed_methods[0]
@@ -381,6 +410,8 @@ class Twilio2FAStartView(Twilio2FAVerificationMixin, TemplateView):
                 self.success_url
             )
 
+        logger.debug("Start dispatch")
+
         return super().dispatch(request, *args, **kwargs)
 
     def get_context_data(self, **kwargs):
@@ -391,7 +422,7 @@ class Twilio2FAStartView(Twilio2FAVerificationMixin, TemplateView):
         return ctx
 
     def retry_action(self, request, *args, **kwargs):
-        elapsed = datetime.now() - datetime.strptime(self.get_session_value(self.SESSION_TIMESTAMP), self.DATEFMT)
+        elapsed = datetime.now() - datetime.strptime(self.get_session_value(SESSION_TIMESTAMP), DATEFMT)
 
         min_retry_wait = get_setting(
             "RETRY_TIME",
@@ -405,7 +436,7 @@ class Twilio2FAStartView(Twilio2FAVerificationMixin, TemplateView):
             )
             return self.get_redirect("verify")
 
-        method = self.get_session_value(self.SESSION_METHOD)
+        method = self.get_session_value(SESSION_METHOD)
 
         self.cancel_verification()
 
@@ -466,17 +497,17 @@ class Twilio2FAStartView(Twilio2FAVerificationMixin, TemplateView):
                 )
             )
 
-            self.set_session_value(self.SESSION_SID, verification.sid)
-            self.set_session_value(self.SESSION_METHOD, method)
-            self.set_session_value(self.SESSION_TIMESTAMP, datetime.now().strftime(self.DATEFMT))
+            self.set_session_value(SESSION_SID, verification.sid)
+            self.set_session_value(SESSION_METHOD, method)
+            self.set_session_value(SESSION_TIMESTAMP, datetime.now().strftime(DATEFMT))
 
             twilio_2fa_verification_sent.send(
-                sender=self.__class__,
-                twilio_sid=self.get_session_value(self.SESSION_SID),
+                sender=None,
+                twilio_sid=self.get_session_value(SESSION_SID),
                 user=self.request.user,
                 phone_number=self.phone_number,
-                method=self.get_session_value(self.SESSION_METHOD),
-                timestamp=self.get_session_value(self.SESSION_TIMESTAMP)
+                method=self.get_session_value(SESSION_METHOD),
+                timestamp=self.get_session_value(SESSION_TIMESTAMP)
             )
 
             return verification.sid
@@ -497,19 +528,36 @@ class Twilio2FAVerifyView(Twilio2FAVerificationMixin, FormView):
     success_url = reverse_lazy("twilio_2fa:success")
     template_name = "twilio_2fa/verify.html"
 
-    def get_success_url(self):
-        return get_setting(
-            "VERIFY_SUCCESS_URL",
-            default=super().get_success_url()
-        )
-
     def get_context_data(self, **kwargs):
         ctx = super().get_context_data(**kwargs)
 
-        ctx["method"] = self.get_session_value(self.SESSION_METHOD)
+        ctx["method"] = self.get_session_value(SESSION_METHOD)
 
         return ctx
 
+    def handle_too_many_attempts(self):
+        self.cancel_verification()
+
+        timeout = get_setting(
+            "MAX_ATTEMPTS_TIMEOUT",
+            default=600
+        )
+
+        if timeout:
+            self.set_session_value(
+                SESSION_TIMEOUT,
+                datetime.strftime(datetime.now() + timedelta(seconds=int(timeout)), DATEFMT)
+            )
+
+        messages.error(
+            self.request,
+            "You have made too many attempts to verify."
+        )
+
+        return self.get_error_redirect(
+            can_retry=True if not timeout else False
+        )
+
     def form_valid(self, form):
         try:
             verify = (get_twilio_client().verify
@@ -523,20 +571,44 @@ class Twilio2FAVerifyView(Twilio2FAVerificationMixin, FormView):
         except TwilioRestException as e:
             if e.code == 60202:
                 # Max tries
-                self.cancel_verification()
-
-                messages.error(
-                    self.request,
-                    "You have made too many attempts to verify."
-                )
-                return self.get_error_redirect(
-                    can_retry=True
-                )
+                return self.handle_too_many_attempts()
 
             return self.handle_twilio_exception(e)
 
+        max_attempts = int(get_setting(
+            "MAX_ATTEMPTS",
+            default=5,
+            callback_kwargs={
+                "user": self.request.user
+            }
+        ))
+
+        current_attempts = self.set_session_value(
+            SESSION_ATTEMPTS,
+            self.get_session_value(SESSION_ATTEMPTS, 0) + 1
+        )
+
+        if current_attempts >= max_attempts:
+            return self.handle_too_many_attempts()
+
         if verify.status == "approved":
-            # TODO: Add signal here
+            # Send this signal manually
+            twilio_2fa_verification_status_changed.send(
+                None,
+                user=self.request.user,
+                phone_number=self.e164_phone_number(),
+                method=self.get_session_value(SESSION_METHOD),
+                twilio_sid=self.get_session_value(SESSION_SID),
+                status=verify.status
+            )
+
+            twilio_2fa_verification_success.send(
+                None,
+                user=self.request.user,
+                phone_number=self.e164_phone_number(),
+                method=self.get_session_value(SESSION_METHOD)
+            )
+
             return super().form_valid(form)
 
         messages.error(
@@ -550,6 +622,27 @@ class Twilio2FAVerifyView(Twilio2FAVerificationMixin, FormView):
 class Twilio2FASuccessView(Twilio2FAMixin, TemplateView):
     template_name = "twilio_2fa/success.html"
 
+    def dispatch(self, request, *args, **kwargs):
+        next_url = self.get_session_value(
+            SESSION_NEXT_URL
+        )
+
+        if next_url:
+            return HttpResponseRedirect(
+                next_url
+            )
+
+        verify_success_url = get_setting(
+            "VERIFY_SUCCESS_URL"
+        )
+
+        if verify_success_url:
+            return HttpResponseRedirect(
+                verify_success_url
+            )
+
+        return super().dispatch(request, *args, **kwargs)
+
 
 class Twilio2FAFailedView(Twilio2FAMixin, TemplateView):
     template_name = "twilio_2fa/failed.html"
@@ -557,7 +650,7 @@ class Twilio2FAFailedView(Twilio2FAMixin, TemplateView):
     def get_context_data(self, **kwargs):
         ctx = super().get_context_data(**kwargs)
 
-        ctx["can_retry"] = self.get_session_value(self.SESSION_CAN_RETRY, False)
+        ctx["can_retry"] = self.get_session_value(SESSION_CAN_RETRY, False)
 
         if settings.DEBUG and "retry" in self.request.GET:
             ctx["can_retry"] = bool(int(self.request.GET.get("retry", 0)))
